package java.util.regex;

import java.util.Objects;

public final class Matcher implements MatchResult
{
    Pattern parentPattern;
    int[] groups;
    int from;
    int to;
    int lookbehindTo;
    CharSequence text;
    static final int ENDANCHOR = 1;
    static final int NOANCHOR = 0;
    int acceptMode;
    int first;
    int last;
    int oldLast;
    int lastAppendPosition;
    int[] locals;
    boolean hitEnd;
    boolean requireEnd;
    boolean transparentBounds;
    boolean anchoringBounds;
    
    Matcher() {
        this.acceptMode = 0;
        this.first = -1;
        this.last = 0;
        this.oldLast = -1;
        this.lastAppendPosition = 0;
        this.transparentBounds = false;
        this.anchoringBounds = true;
    }
    
    Matcher(Pattern parentPattern, CharSequence text) {
        this.acceptMode = 0;
        this.first = -1;
        this.last = 0;
        this.oldLast = -1;
        this.lastAppendPosition = 0;
        this.transparentBounds = false;
        this.anchoringBounds = true;
        this.parentPattern = parentPattern;
        this.text = text;
        this.groups = new int[Math.max(parentPattern.capturingGroupCount, 10) * 2];
        this.locals = new int[parentPattern.localCount];
        this.reset();
    }
    
    public Pattern pattern() {
        return this.parentPattern;
    }
    
    public MatchResult toMatchResult() {
        Matcher matcher = new Matcher(this.parentPattern, this.text.toString());
        matcher.first = this.first;
        matcher.last = this.last;
        matcher.groups = (int[])this.groups.clone();
        return matcher;
    }
    
    public Matcher usePattern(Pattern parentPattern) {
        if (parentPattern == null) {
            throw new IllegalArgumentException("Pattern cannot be null");
        }
        this.parentPattern = parentPattern;
        this.groups = new int[Math.max(parentPattern.capturingGroupCount, 10) * 2];
        this.locals = new int[parentPattern.localCount];
        for (int i = 0; i < this.groups.length; ++i) {
            this.groups[i] = -1;
        }
        for (int j = 0; j < this.locals.length; ++j) {
            this.locals[j] = -1;
        }
        return this;
    }
    
    public Matcher reset() {
        this.first = -1;
        this.last = 0;
        this.oldLast = -1;
        for (int i = 0; i < this.groups.length; ++i) {
            this.groups[i] = -1;
        }
        for (int j = 0; j < this.locals.length; ++j) {
            this.locals[j] = -1;
        }
        this.lastAppendPosition = 0;
        this.from = 0;
        this.to = this.getTextLength();
        return this;
    }
    
    public Matcher reset(CharSequence text) {
        this.text = text;
        return this.reset();
    }
    
    @Override
    public int start() {
        if (this.first < 0) {
            throw new IllegalStateException("No match available");
        }
        return this.first;
    }
    
    @Override
    public int start(int i) {
        if (this.first < 0) {
            throw new IllegalStateException("No match available");
        }
        if (i < 0 || i > this.groupCount()) {
            throw new IndexOutOfBoundsException("No group " + i);
        }
        return this.groups[i * 2];
    }
    
    public int start(String s) {
        return this.groups[this.getMatchedGroupIndex(s) * 2];
    }
    
    @Override
    public int end() {
        if (this.first < 0) {
            throw new IllegalStateException("No match available");
        }
        return this.last;
    }
    
    @Override
    public int end(int i) {
        if (this.first < 0) {
            throw new IllegalStateException("No match available");
        }
        if (i < 0 || i > this.groupCount()) {
            throw new IndexOutOfBoundsException("No group " + i);
        }
        return this.groups[i * 2 + 1];
    }
    
    public int end(String s) {
        return this.groups[this.getMatchedGroupIndex(s) * 2 + 1];
    }
    
    @Override
    public String group() {
        return this.group(0);
    }
    
    @Override
    public String group(int i) {
        if (this.first < 0) {
            throw new IllegalStateException("No match found");
        }
        if (i < 0 || i > this.groupCount()) {
            throw new IndexOutOfBoundsException("No group " + i);
        }
        if (this.groups[i * 2] == -1 || this.groups[i * 2 + 1] == -1) {
            return null;
        }
        return this.getSubSequence(this.groups[i * 2], this.groups[i * 2 + 1]).toString();
    }
    
    public String group(String s) {
        int matchedGroupIndex = this.getMatchedGroupIndex(s);
        if (this.groups[matchedGroupIndex * 2] == -1 || this.groups[matchedGroupIndex * 2 + 1] == -1) {
            return null;
        }
        return this.getSubSequence(this.groups[matchedGroupIndex * 2], this.groups[matchedGroupIndex * 2 + 1]).toString();
    }
    
    @Override
    public int groupCount() {
        return this.parentPattern.capturingGroupCount - 1;
    }
    
    public boolean matches() {
        return this.match(this.from, 1);
    }
    
    public boolean find() {
        int n = this.last;
        if (n == this.first) {
            ++n;
        }
        if (n < this.from) {
            n = this.from;
        }
        if (n > this.to) {
            for (int i = 0; i < this.groups.length; ++i) {
                this.groups[i] = -1;
            }
            return false;
        }
        return this.search(n);
    }
    
    public boolean find(int n) {
        int textLength = this.getTextLength();
        if (n < 0 || n > textLength) {
            throw new IndexOutOfBoundsException("Illegal start index");
        }
        this.reset();
        return this.search(n);
    }
    
    public boolean lookingAt() {
        return this.match(this.from, 0);
    }
    
    public static String quoteReplacement(String s) {
        if (s.indexOf(92) == -1 && s.indexOf(36) == -1) {
            return s;
        }
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < s.length(); ++i) {
            int char1 = s.charAt(i);
            if (char1 == '\\' || char1 == '$') {
                sb.append('\\');
            }
            sb.append((char)char1);
        }
        return sb.toString();
    }
    
    public Matcher appendReplacement(StringBuffer sb, String s) {
        if (this.first < 0) {
            throw new IllegalStateException("No match available");
        }
        int i = 0;
        StringBuilder s2 = new StringBuilder();
        while (i < s.length()) {
            int char1 = s.charAt(i);
            if (char1 == '\\') {
                if (++i == s.length()) {
                    throw new IllegalArgumentException("character to be escaped is missing");
                }
                s2.append(s.charAt(i));
                ++i;
            }
            else if (char1 == '$') {
                if (++i == s.length()) {
                    throw new IllegalArgumentException("Illegal group reference: group index is missing");
                }
                int c = s.charAt(i);
                int intValue;
                if (c == '{') {
                    ++i;
                    StringBuilder o = new StringBuilder();
                    while (i < s.length()) {
                        c = s.charAt(i);
                        if (ASCII.isLower(c) || ASCII.isUpper(c) || ASCII.isDigit(c)) {
                            o.append((char)c);
                            ++i;
                        }
                        else {
                            break;
                        }
                    }
                    if (o.length() == 0) {
                        throw new IllegalArgumentException("named capturing group has 0 length name");
                    }
                    if (c != '}') {
                        throw new IllegalArgumentException("named capturing group is missing trailing '}'");
                    }
                    String string = o.toString();
                    if (ASCII.isDigit((int)string.charAt(0))) {
                        throw new IllegalArgumentException("capturing group name {" + string + "} starts with digit character");
                    }
                    if (!this.parentPattern.namedGroups().containsKey(string)) {
                        throw new IllegalArgumentException("No group with name {" + string + "}");
                    }
                    intValue = (int)this.parentPattern.namedGroups().get(string);
                    ++i;
                }
                else {
                    intValue = c - 48;
                    if (intValue < 0 || intValue > 9) {
                        throw new IllegalArgumentException("Illegal group reference");
                    }
                    ++i;
                    int j = 0;
                    while (j == 0) {
                        if (i >= s.length()) {
                            break;
                        }
                        int n = s.charAt(i) - 48;
                        if (n >= 0) {
                            if (n > 9) {
                                break;
                            }
                            int n2 = intValue * 10 + n;
                            if (this.groupCount() < n2) {
                                j = 1;
                            }
                            else {
                                intValue = n2;
                                ++i;
                            }
                        }
                        else {
                            break;
                        }
                    }
                }
                if (this.start(intValue) != -1 && this.end(intValue) != -1) {
                    s2.append(this.text, this.start(intValue), this.end(intValue));
                }
                else {
                    continue;
                }
            }
            else {
                s2.append((char)char1);
                ++i;
            }
        }
        sb.append(this.text, this.lastAppendPosition, this.first);
        sb.append(s2);
        this.lastAppendPosition = this.last;
        return this;
    }
    
    public StringBuffer appendTail(StringBuffer sb) {
        sb.append(this.text, this.lastAppendPosition, this.getTextLength());
        return sb;
    }
    
    public String replaceAll(String s) {
        this.reset();
        if (this.find()) {
            StringBuffer sb = new StringBuffer();
            do {
                this.appendReplacement(sb, s);
            } while (this.find());
            this.appendTail(sb);
            return sb.toString();
        }
        return this.text.toString();
    }
    
    public String replaceFirst(String s) {
        if (s == null) {
            throw new NullPointerException("replacement");
        }
        this.reset();
        if (!this.find()) {
            return this.text.toString();
        }
        StringBuffer sb = new StringBuffer();
        this.appendReplacement(sb, s);
        this.appendTail(sb);
        return sb.toString();
    }
    
    public Matcher region(int from, int to) {
        if (from < 0 || from > this.getTextLength()) {
            throw new IndexOutOfBoundsException("start");
        }
        if (to < 0 || to > this.getTextLength()) {
            throw new IndexOutOfBoundsException("end");
        }
        if (from > to) {
            throw new IndexOutOfBoundsException("start > end");
        }
        this.reset();
        this.from = from;
        this.to = to;
        return this;
    }
    
    public int regionStart() {
        return this.from;
    }
    
    public int regionEnd() {
        return this.to;
    }
    
    public boolean hasTransparentBounds() {
        return this.transparentBounds;
    }
    
    public Matcher useTransparentBounds(boolean transparentBounds) {
        this.transparentBounds = transparentBounds;
        return this;
    }
    
    public boolean hasAnchoringBounds() {
        return this.anchoringBounds;
    }
    
    public Matcher useAnchoringBounds(boolean anchoringBounds) {
        this.anchoringBounds = anchoringBounds;
        return this;
    }
    
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("java.util.regex.Matcher");
        sb.append("[pattern=" + this.pattern());
        sb.append(" region=");
        sb.append(this.regionStart() + "," + this.regionEnd());
        sb.append(" lastmatch=");
        if (this.first >= 0 && this.group() != null) {
            sb.append(this.group());
        }
        sb.append("]");
        return sb.toString();
    }
    
    public boolean hitEnd() {
        return this.hitEnd;
    }
    
    public boolean requireEnd() {
        return this.requireEnd;
    }
    
    boolean search(int first) {
        this.hitEnd = false;
        this.requireEnd = false;
        first = ((first < 0) ? 0 : first);
        this.first = first;
        this.oldLast = ((this.oldLast < 0) ? first : this.oldLast);
        for (int i = 0; i < this.groups.length; ++i) {
            this.groups[i] = -1;
        }
        this.acceptMode = 0;
        int match = this.parentPattern.root.match(this, first, this.text);
        if (match == 0) {
            this.first = -1;
        }
        this.oldLast = this.last;
        return match;
    }
    
    boolean match(int first, int acceptMode) {
        this.hitEnd = false;
        this.requireEnd = false;
        first = ((first < 0) ? 0 : first);
        this.first = first;
        this.oldLast = ((this.oldLast < 0) ? first : this.oldLast);
        for (int i = 0; i < this.groups.length; ++i) {
            this.groups[i] = -1;
        }
        this.acceptMode = acceptMode;
        int match = this.parentPattern.matchRoot.match(this, first, this.text);
        if (match == 0) {
            this.first = -1;
        }
        this.oldLast = this.last;
        return match;
    }
    
    int getTextLength() {
        return this.text.length();
    }
    
    CharSequence getSubSequence(int n, int n2) {
        return this.text.subSequence(n, n2);
    }
    
    char charAt(int n) {
        return this.text.charAt(n);
    }
    
    int getMatchedGroupIndex(String s) {
        Objects.requireNonNull(s, "Group name");
        if (this.first < 0) {
            throw new IllegalStateException("No match found");
        }
        if (!this.parentPattern.namedGroups().containsKey(s)) {
            throw new IllegalArgumentException("No group with name <" + s + ">");
        }
        return (int)this.parentPattern.namedGroups().get(s);
    }
}
